#include "pop_pal_fast.h"

// converts to 8 bit using palette
// - optimisation here is to only use the slow 
// GetNearestPaletteIndex once for each colour
// used in the bitmap - and also to initialise
// the colour space with the palette entries
// so those don't need to be searched for at all.
// Then use a sparse array for the colour space
// to reduce the memory allocation

// I call the freeimage routines to find the
// scan lines and colour organisation etc - 
// as that is what I use in the original program
// - it is simpler to use than the Windows
// routines to find the pixel values. To use those you would find a
// device independent bitmap first and then
// use that instead of the FIBITMAP * arguments
// You get it using GetDIBits(hdc, hBitmap,...)
// to make a device dependent bitmap from
// one created for the current display - see
// chapter 15 seciton II of Charles Petzold's
// programming windows. Don't use GetPixelValue(hdc,x,y)
// in production code - it is very slow indeed.

int **ppiColourSpace;
DWORD dwMemColourSpace;
double dMaxMemColourSpace;

int ires_colour_space=8;
#define RES_SHIFT 6
#define FIND_CLR(irgb) (ppiColourSpace[irgb>>RES_SHIFT]?ppiColourSpace[irgb>>RES_SHIFT][irgb%(1<<RES_SHIFT)]:0)
#define SET_CLR(irgb,val)\
{\
  if(!ppiColourSpace[irgb>>RES_SHIFT])\
  {ppiColourSpace[irgb>>RES_SHIFT]=calloc(sizeof(int),(1<<RES_SHIFT));\
   dwMemColourSpace+=64*sizeof(int);\
  }\
  if(ppiColourSpace[irgb>>RES_SHIFT])\
   ppiColourSpace[irgb>>RES_SHIFT][irgb%(1<<RES_SHIFT)]=val;\
}

int GetNearestPaletteIndexOptimised(HPALETTE hPalette,COLORREF clrf)
{
 if(ppiColourSpace)
 {
  BYTE r=GetRValue(clrf)>>(8-ires_colour_space),g=GetGValue(clrf)>>(8-ires_colour_space),b=GetBValue(clrf)>>(8-ires_colour_space);
  int irgb=(r<<(ires_colour_space*2))+(g<<ires_colour_space)+b;
  int iret=FIND_CLR(irgb)-1;
  if(iret<0)
  {
   iret=GetNearestPaletteIndex(hPalette,clrf);
   SET_CLR(irgb,iret+1);
  }
  return iret;
 }
 return (int)GetNearestPaletteIndex(hPalette,clrf);
}

#define RGB_SHIFT3(r,g,b) RGB((r)<<3,(g)<<3,(b)<<3)
void ConvertLine16To8_Colour(HPALETTE hPalette,BYTE *target, BYTE *source, int width_in_pixels) 
{
 int cols =0;
	WORD *bits = (WORD *)source;
	for (cols = 0; cols < width_in_pixels; cols++)
 {
		target[cols] =GetNearestPaletteIndexOptimised
   (hPalette,RGB_SHIFT3 ((bits[cols] & 0x7C00) >> 10,(bits[cols] & 0x3E0) >> 5, (bits[cols] & 0x1F)));
	}
}

void ConvertLine24To8_Colour(HPALETTE hPalette,BYTE *target, BYTE *source, int width_in_pixels) 
{
 int cols =0;
	for ( cols = 0; cols < width_in_pixels; cols++)
 {
		target[cols] = GetNearestPaletteIndexOptimised
   (hPalette,RGB (source[2], source[1], source[0]));
		source += 3;
	}
}


void ConvertLine32To8_Colour(HPALETTE hPalette,BYTE *target, BYTE *source, int width_in_pixels) 
{
 int cols =0;
	for (cols = 0; cols < width_in_pixels; cols++) 
 {
		target[cols] = GetNearestPaletteIndexOptimised
   (hPalette,RGB (source[2], source[1], source[0]));
		source += 4;
	}
}

FIBITMAP *ConvertTo8BitsWithPalette( HPALETTE hPalette,FIBITMAP *dib,int ires)
{
 // Note. This allocates a new FIBITMAP but doesn't free the old one
 // - be sure to free the old one in calling proc after conversion or
 // you will get a memory leak
 if(dib)
 {
	 if(dib&&FreeImage_GetBPP(dib)<8)
   return FreeImage_ConvertTo8Bits(dib);
  if(dib&&FreeImage_GetBPP(dib)==8)
   return FreeImage_Clone(dib);
 }
 if(dib)
 {
  int rows=0;
		int width  = FreeImage_GetWidth(dib);
		int height = FreeImage_GetHeight(dib);
		FIBITMAP *new_dib = FreeImage_Allocate(width, height, 8,0,0,0);
  WORD wNumEntries;
  PALETTEENTRY pe ;
  int imax_ppiColourSpace=0;
  int i=0;
  RGBQUAD *new_pal;
  ires_colour_space=max(1,min(8,ires));
		if (new_dib == NULL)
			return NULL;
 	new_pal = FreeImage_GetPalette(new_dib);
  // this is our sparse array for the colour space -
  // each element of the sparse array indexes into (1<<RES_SHIFT) consecutive
  // entries of the colour space - and these only get allocated if 
  // one of the entries in the range needs to be set
  imax_ppiColourSpace=max(1,(1<<(ires_colour_space*3))>>RES_SHIFT);
  ppiColourSpace=(int **)calloc(sizeof(int *),imax_ppiColourSpace);

  GetObject (hPalette, sizeof (WORD), &wNumEntries) ;
  for (i = 0 ; i < (int) wNumEntries &&i<255; i++)
  {
   GetPaletteEntries (hPalette, i, 1, &pe) ;

   new_pal[i].rgbRed   = pe.peRed ;
   new_pal[i].rgbGreen = pe.peGreen ;
   new_pal[i].rgbBlue  = pe.peBlue ;
   new_pal[i].rgbReserved = 0 ;
   // record these indexes in our colour for pixel value
   // look up table - saves a number of calls to the slow GetNearestPaletteIndex(...)
   // in fact if the original bitmap has few colours, enough for all of them
   // to be included in the palette, then GetNearestPaletteIndex may not get called at all
   {
    BYTE r=pe.peRed>>(8-ires_colour_space),g=pe.peGreen>>(8-ires_colour_space),b=pe.peBlue>>(8-ires_colour_space);
    int irgb=(r<<(ires_colour_space*2))+(g<<ires_colour_space)+b;
    SET_CLR(irgb,i+1)
   }
  }
		switch(FreeImage_GetBPP(dib))
  {
			case 16 :
			{
				for (rows = 0; rows < height; rows++) {
						ConvertLine16To8_Colour(hPalette,FreeImage_GetScanLine(new_dib, rows), FreeImage_GetScanLine(dib, rows), width);
    if(GetAsyncKeyState(VK_ESCAPE)<0)
     break;
					}
				}
			break;
			case 24 :
			{
				for (rows = 0; rows < height; rows++)
    {
					ConvertLine24To8_Colour(hPalette,FreeImage_GetScanLine(new_dib, rows), FreeImage_GetScanLine(dib, rows), width);					
    if(GetAsyncKeyState(VK_ESCAPE)<0)
     break;
    }

			}
			break;
  	case 32 :
			{
				for (rows = 0; rows < height; rows++)
    {
					ConvertLine32To8_Colour(hPalette,FreeImage_GetScanLine(new_dib, rows), FreeImage_GetScanLine(dib, rows), width);
    if(GetAsyncKeyState(VK_ESCAPE)<0)
     break;
					}
				
			}
  	break;
		}
  if(ppiColourSpace)
  {
   int i=0;
   for(i=0;i<imax_ppiColourSpace;i++)
   FREE_IF_NON_ZERO(ppiColourSpace[i])
  }
  FREE_IF_NON_ZERO(ppiColourSpace);
  {
   double dMem=(double)dwMemColourSpace/(1024.0*1024.0);
   dMaxMemColourSpace=max(dMem,dMaxMemColourSpace);
  }
 	return new_dib;
	}
 return NULL;
}

FIBITMAP *ConvertTo8BitsColourDestroyOriginal(FIBITMAP *dib)
{
 // Note. This destroys the old bitmap 
 // so the original FIBITMAP handle is no longer valid on return
 // - if unsuccessful it still destroys it and returns NULL
 // - on the assumption that if it fails then you don't want
 // - to do anything with the bitmap anyway.
 // - 
 // - typical use:
 // dib=FreeImage_ConvertTo8BitsColourDestroyOriginal(dib);
 // if(dib)
 // {...
 // }
 double dMem=0;
 int iMaxRes=8;
 FIBITMAP *new_dib=NULL;
 if(dib)
 {
  HPALETTE hPalette=PopPalFast(dib,&iMaxRes,&dMem);
  new_dib=ConvertTo8BitsWithPalette(hPalette,dib,iMaxRes);
  FreeImage_Unload(dib);
 }
 return new_dib;
}

FIBITMAP *ConvertTo8BitsColour(FIBITMAP *dib)
{
 // Note. This allocates a new FIBITMAP but doesn't free the old one
 // - be sure to free the old one in calling proc after conversion or
 // you will get a memory leak
 double dMem=0;
 int iMaxRes=8;
 HPALETTE hPalette=PopPalFast(dib,&iMaxRes,&dMem);
 return ConvertTo8BitsWithPalette(hPalette,dib,iMaxRes);
}
