#include "pop_pal_fast.h"

// Speeded up popularity algorithm - used the unoptimised code in 
// Charles Petzold Programming Windows.as a starting point.

// This super fast popularity algorithm uses sparse arrays and various optimisations
// Robert Walker support@tunesmithy.co.uk
// home page http://tunesmithy.co.uk/poppalfast/index.htm

// The routine is 

HPALETTE PopPalFast(FIBITMAP *dib, int *piMaxRes,double *pdMem);

// Call it with your bitmap and the desired maximum resolution which
// could be 8 for best possble unless you particularly want a faster 
// lower resolution. 

// Returns the new palette.

// There is no particular advantage in calling it
// with low resolution as it goes through the palettes in increasing
// order of size anyway and stops when it finds enough palette entries
// - then it uses the previously found palette and extends it to a 
// complete 236 colour palette by adding any of the new ones at maximum
// possible colour distance from ones so far in it.
// 
// On return, *piMaxRes gets set to the actual resolution used
// and *pdMem gets set to the amount of memory that got allocated
// to do the counts - expect this to normally be pretty small
// - less than 1 Mb anyway, because of the sparse arrays.
// Set pdMem to NULL if you don't need this info

// Note about the FIBITMAP argument here

// I call the FreeImage routines to find the
// scan lines and colour organisation etc - 
// as that is what I use in the original program
// - it is simpler to use than the Windows
// routines to find the pixel values.
// and it will also focus attention on the
// implementation of the algorithm as the
// FreeImage syntax is really clear to read. 
//
// To use Windows routiens you might perhaps find a
// device independent bitmap first and then
// use that instead of the FIBITMAP * arguments
// You get it using GetDIBits(hdc, hBitmap,...)
// to make a device dependent bitmap from
// one created for the current display - for details see
// chapter 15 seciton II of Charles Petzold's
// programming windows. 
//
// Don't use  
// COLORREF cr=GetPixel(hdcMem, cx, cy) ;
// in production code - it is very slow indeed.
// and if you use a fast palette construction method
// your program will spend most of its time
// on that line.

/*--------------------------------------------------------------------
   AccumColorCounts: Fills up ppiCount (indexed by a packed RGB color)
     with counts of pixels of that color.
  --------------------------------------------------------------------*/
#define COUNT_SHIFT 6 // how much to shift to make the sparse array
                      // - each array pointer in the sparse array addresses
                      // a row of (1<<COUNT_SHIFT) colour values
                      // if it gets used


#define PACK_RGB(R,G,B,iRes) ((int) (R) | ((int) (G) <<  (iRes)) |       \
                                          ((int) (B) << ((iRes) + (iRes))))

static double dMaxMemCount;// for inspection with debugger

// this is much faster as a macro when used for the unoptimised debug code.

#define ACCUM_COUNT(R,G,B,ppiCount,ppiOrig,iRes,pdwMem)\
{\
 BYTE r=R,g=G,b=B;\
 r  >>= (8 - iRes) ;\
 g >>= (8 - iRes) ;\
 b  >>= (8 - iRes) ;\
 {\
  int irgb=PACK_RGB (r,g,b, iRes);\
  if(!ppiCount[irgb>>COUNT_SHIFT])\
  {ppiCount[irgb>>COUNT_SHIFT]=calloc(sizeof(int),(1<<COUNT_SHIFT));\
   (*pdwMem)+=64*sizeof(int);\
   if(ppiOrig)ppiOrig[irgb>>COUNT_SHIFT]=calloc(sizeof(int),(1<<COUNT_SHIFT));\
  }\
  if(ppiCount[irgb>>COUNT_SHIFT])\
   ppiCount[irgb>>COUNT_SHIFT][irgb%(1<<COUNT_SHIFT)]++;\
  if(ppiOrig&&ppiOrig[irgb>>COUNT_SHIFT]&&!ppiOrig[irgb>>COUNT_SHIFT][irgb%(1<<COUNT_SHIFT)])\
  {\
   int irgb_orig=PACK_RGB (R,G,B, 8);\
   ppiOrig[irgb>>COUNT_SHIFT][irgb%(1<<COUNT_SHIFT)]=irgb_orig;\
  }\
 }\
}
#define FIND_COUNT(irgb) (ppiCount[irgb>>COUNT_SHIFT]?ppiCount[irgb>>COUNT_SHIFT][irgb%(1<<COUNT_SHIFT)]:0)
#define FIND_ORIG(irgb) (ppiOrig&&ppiOrig[irgb>>COUNT_SHIFT]?ppiOrig[irgb>>COUNT_SHIFT][irgb%(1<<COUNT_SHIFT)]:irgb)
#define ACCUM_COUNT_SHIFT_3(R,G,B,ppiCount,ppiOrig,iRes,pdwMem)\
 ACCUM_COUNT((R)<<3,(G)<<3,(B)<<3,ppiCount,ppiOrig,iRes,pdwMem)

void AccumLineColours16_565(BYTE *source, int width_in_pixels
                                       , int * * ppiCount, int * * ppiOrig, int iRes, DWORD *pdwMem)
{
 int cols =0;
	WORD *bits = (WORD *)source;
	for (cols = 0; cols < width_in_pixels; cols++)
 {
  BYTE R=(BYTE)( (bits[cols] & (BYTE)(0x1f<<11)) >> 11);
  BYTE G=(BYTE)( (bits[cols] & (BYTE)(0x3f<<5)) >> 5);
  BYTE B=(BYTE)( bits[cols] & (BYTE)(0x1F) );
		ACCUM_COUNT_SHIFT_3(R,G,B,ppiCount,ppiOrig,iRes, pdwMem)
	}
}

void AccumLineColours16(BYTE *source, int width_in_pixels
                                       , int * * ppiCount, int * * ppiOrig, int iRes, DWORD *pdwMem)
{
 int cols =0;
	WORD *bits = (WORD *)source;
	for (cols = 0; cols < width_in_pixels; cols++)
 {
  BYTE R=(BYTE)( (bits[cols] & (BYTE)(0x1f<<10)) >> 10);
  BYTE G=(BYTE)( (bits[cols] & (BYTE)(0x1f<<5)) >> 5);
  BYTE B=(BYTE)( bits[cols] & (BYTE)(0x1F));
		ACCUM_COUNT_SHIFT_3(R,G,B,ppiCount,ppiOrig,iRes, pdwMem)
	}
}

void AccumLineColours24(BYTE *source, int width_in_pixels
                                       , int * * ppiCount, int * * ppiOrig, int iRes, DWORD *pdwMem)
{
 int cols =0;
	for ( cols = 0; cols < width_in_pixels; cols++) 
 {
		ACCUM_COUNT(source[2], source[1], source[0], ppiCount, ppiOrig, iRes, pdwMem)
		source += 3;
	}
}


void AccumLineColours32(BYTE *source, int width_in_pixels
                                  ,int * * ppiCount, int * * ppiOrig, int iRes, DWORD *pdwMem) 
{
 int cols =0;
	for (cols = 0; cols < width_in_pixels; cols++)
 {
		ACCUM_COUNT(source[2], source[1], source[0], ppiCount, ppiOrig, iRes, pdwMem)
		source += 4;
	}
}

static void AccumColorCounts(FIBITMAP *dib, int * * ppiCount, int * * ppiOrig, int iRes, DWORD *pdwMem)
{
  int rows=0;
		int width  = FreeImage_GetWidth(dib);
		int height = FreeImage_GetHeight(dib);
 	switch(FreeImage_GetBPP(dib))
  {
			case 16 :
   {
    BYTE rMask=FreeImage_GetRedMask(dib);
    BYTE gMask=FreeImage_GetGreenMask(dib);
    BYTE bMask=FreeImage_GetBlueMask(dib);
    if(rMask==0x1F&&gMask==0x3F<<5&&bMask==0x1F<<11)
				for (rows = 0; rows < height; rows++) 
						AccumLineColours16_565(FreeImage_GetScanLine(dib, rows), width, ppiCount, ppiOrig,iRes, pdwMem);
    else
				for (rows = 0; rows < height; rows++) 
						AccumLineColours16(FreeImage_GetScanLine(dib, rows), width, ppiCount, ppiOrig,iRes, pdwMem);
   }
				return;
			case 24 :
				for (rows = 0; rows < height; rows++)
						AccumLineColours24(FreeImage_GetScanLine(dib, rows), width, ppiCount, ppiOrig,iRes, pdwMem);
    return;
			case 32 :
				for (rows = 0; rows < height; rows++)
						AccumLineColours32(FreeImage_GetScanLine(dib, rows), width, ppiCount, ppiOrig,iRes, pdwMem);
    return;
		}
}

/*--------------------------------------------------------------
   DibPalPopularity:  Popularity algorithm for optimized colors
  --------------------------------------------------------------*/
HPALETTE PopPalFast(FIBITMAP *dib, int *piMaxRes,double *pdMem)
{
 int iMaxRes=*piMaxRes;// *piMaxRes gets set to the actual highest resolution used on return
 HPALETTE     hPaletteIs=NULL;
 LOGPALETTE * plp=NULL,*plp_was=NULL;
 int     *  * ppiCount=NULL;
 int     *  * ppiOrig=NULL;// this keeps track of an original pixel value for each
                           // truncated pixel counted
 int iEntryWas=0;
 int iRes=0,iEntry=0, iArraySize=0;
 if(!dib)
  return NULL;
 iMaxRes=max(iMaxRes,COUNT_SHIFT);
 // Allocate array for counting pixel colors
 // iRes here is the number of bits to use for each colour value
 // up to iRes==8 means use 8 bit 
 for(iRes=COUNT_SHIFT/3;iRes<=iMaxRes;iRes++)
 {
  int  i=0,  iCount=0, iIndex=0, iMask=0, R=0, G=0, B=0 ;
  DWORD dwMem=0;
  iEntry=0;
  if(ppiCount)
  {
   for(i=0;i<iArraySize;i++)
   FREE_IF_NON_ZERO(ppiCount[i])
  }
  FREE_IF_NON_ZERO(ppiCount)
  if(ppiOrig)
  {
   for(i=0;i<iArraySize;i++)
   FREE_IF_NON_ZERO(ppiOrig[i])
  }
  FREE_IF_NON_ZERO(ppiOrig)
  plp=NULL;
  iArraySize = (1 << (3 * iRes)) >>COUNT_SHIFT;
  // iMask = (1 << iRes) - 1 ;
  iMask = (1 << 8) - 1 ; 
  // now uses a colour of one of the original pixels for each bin for the
  // resolution reached
  if(GetAsyncKeyState(VK_ESCAPE)<0)
   break;
  ppiCount=(int **)calloc (iArraySize,sizeof (int *));
  ppiOrig=(int **)calloc (iArraySize,sizeof (int *));
  dwMem=iArraySize*2*sizeof (int *);
  if(ppiCount==NULL)
   break;
       // Get the color counts
  AccumColorCounts(dib, ppiCount,ppiOrig,iRes,&dwMem) ;
       // Set up a palette
  if(pdMem)
  {
   *pdMem=(double)(dwMem)/(1024.0*1024.0);
  }
  {
   int ipos=0;
#define MAX_BEST 255
   int Best[MAX_BEST];
   int Indexes[MAX_BEST];
   int min_best=MAX_BEST-1;
   int i,j,k;
   int icount_shifted;
   Best[min_best]=0;
   Indexes[min_best]=0;
   plp = malloc (sizeof (LOGPALETTE) + 255 * sizeof (PALETTEENTRY)) ;
   if(!plp)
    break;
   for(i=0;i<MAX_BEST;i++)
    Best[i]=Indexes[i]=0;
   // do a sort of just the top 255 entries.(could do more than this
   // - later on after this loop, the ones furthest away from already existing entries in 
   // the palette are the ones that get used, up to 236 only
   for (icount_shifted = 0; icount_shifted < iArraySize ; icount_shifted++)
   {
    int icount_pos=0;
    if(ppiCount[icount_shifted])
    for(icount_pos=0;icount_pos<(1<<COUNT_SHIFT);icount_pos++)
    {
     int index=icount_shifted*(1<<COUNT_SHIFT)+icount_pos;
     int inew_count=ppiCount[icount_shifted][icount_pos];
     if(inew_count>0)
     if(min_best>0||inew_count>Best[0])
     {
      int jnew=0;
      // this test should fail often as one gets towards
      // the end of the search so this will actually be quite
      // fast probably
      for(j=max(min_best,1);j<MAX_BEST;j++)
       if(Best[j]>inew_count||(min_best>0&&Indexes[j]>index))// keep indexes in increasing order in each category
        break;
       jnew=j-1;
      min_best=max(0,min_best-1);
      // insert new entry before this one and move all the earlier ones down one
      // - this could probably be optimised better by using some kind of a binary
      // tree - but it is already pretty fast as it is. One needs to be sure that 
      // one can check if a value is less than the smallest value so far in the 
      // tree very speedily, as that is what is most important with bitmaps
      // which have many more than 256 colours in them. This simple method is
      // exceedingly fast for such bitmaps as usually you don't need to
      // add new entries at all.
      for(k=min_best;k<jnew;k++)
      {
       Best[k]=Best[k+1];
       Indexes[k]=Indexes[k+1];
      }
      Best[jnew]=inew_count;
      Indexes[jnew]=index;
      min_best=min(min_best,jnew);
      if(min_best==0)
       min_best=min_best;
     }
    }
   }
   for (i = MAX_BEST-1 ; i >=min_best; i--)
   {
    iEntry=MAX_BEST-1-i;
    iCount=Best[i];
    iIndex=FIND_ORIG(Indexes[i]);
    if (iCount == 0)
      break ;
        R = (iMask &  iIndex     ) ;
        G = (iMask & (iIndex >> 8)) ;
        B = (iMask & (iIndex >>16))  ;
    plp->palPalEntry[iEntry].peRed   = (BYTE) R ; 
    plp->palPalEntry[iEntry].peGreen = (BYTE) G ; 
    plp->palPalEntry[iEntry].peBlue  = (BYTE) B ; 
    plp->palPalEntry[iEntry].peFlags = 0 ;
   }
   iEntry++;
   // On exit from the loop iEntry will be the number of stored entries

   }
  // Use the first found palette with more than 236 entries
  if(iRes>0&&iEntry>=236)
  {
   break;
  }
  /**
  Note - if increasing the resolution finds no new
  // colours then it will often mean that we are done.
  // so one could have an extra check here for that here - 
  // - but there may be new colours very close to ones already 
  // found which will only show up two resolutions or more afterwards
  //  so best to continue even if this 
  // level of resolution found no new colours
  ***/
  else
  {
   if(iRes<iMaxRes)
   {
    FREE_IF_NON_ZERO(plp_was);
    iEntryWas=iEntry;
    plp_was=plp;
   }
  }
 }
 iRes=min(iRes,iMaxRes);// needed if it goes all the way through the loop
 // and doesn't find any more than 255 entries
 if(plp==NULL&&plp_was)
 {
  plp=plp_was;// case where allocation of array failed for larger arrays
  iEntry=iEntryWas;
  plp_was=NULL;
 }
 if(plp&&plp_was)
 {
  int iwas=0;
  int ipos_new=0;
  // use the entries in plp_was
  // then augment by as many as possible of the new entries in plp 
  LOGPALETTE * plp_new = malloc (sizeof (LOGPALETTE) + 255 * sizeof (PALETTEENTRY)) ;
  if(plp_new)
  {
   int nloops=0;
   for(iwas=0;iwas<iEntryWas;iwas++)
   {
    plp_new->palPalEntry[ipos_new++]=plp_was->palPalEntry[iwas];
   }
 #define SAME_PALETTE_ENTRY(a,b)\
    (a.peRed==b.peRed&&a.peGreen==b.peGreen&&a.peBlue==b.peBlue)
   for(nloops=0;nloops<1000;nloops++)
     // nloops shouldnt be needed here - just an extra runaway loop safeguard
   {
     // Find the entry in the new palette most distant from all the
     // ones found so far
    {
     int i=0;
     int isofarnew=0;
     int max_min_dist=0,ipos_max_min_dist=0;
  #define PALETTE_DISTANCE(a,b)\
    (abs(a.peRed-b.peRed)+abs(a.peGreen-b.peGreen)+abs(a.peBlue-b.peBlue))
     for(i=0;i<iEntry;i++)
     {
      int min_dist=1<<30;
      for(isofarnew=0;isofarnew<ipos_new;isofarnew++)
      {
       int dist=PALETTE_DISTANCE(plp_new->palPalEntry[isofarnew],plp->palPalEntry[i]);
       if(dist<min_dist)
       {
        min_dist=dist;
        if(min_dist==0)
         break;
       }
      }
      if(min_dist>max_min_dist)
      {
       max_min_dist=min_dist;
       ipos_max_min_dist=i;
      }
     }
     if(max_min_dist>0)
     {
      plp_new->palPalEntry[ipos_new++]=plp->palPalEntry[ipos_max_min_dist];
     }
     else
      // all remaining entries are identical to ones already in the palette
      goto finished_palette_lookup;
     if(ipos_new>=236)
      goto finished_palette_lookup;
    }
   }
   finished_palette_lookup:;
   iEntry=ipos_new;
   free(plp);
   plp=plp_new;
  }
 }
 if(plp)
 {
  plp->palVersion = 0x0300 ;
  plp->palNumEntries=min(236,iEntry);
  // Create the palette, clean up, and return the palette handle
    hPaletteIs = CreatePalette (plp) ;
     {
      WORD wNumEntries;
      GetObject (hPaletteIs, sizeof (WORD), &wNumEntries) ;
     }
 }
 FREE_IF_NON_ZERO (plp) ;
 FREE_IF_NON_ZERO (plp_was) ;
 if(ppiOrig)
 {
  int i=0;
  for(i=0;i<iArraySize;i++)
  FREE_IF_NON_ZERO(ppiOrig[i])
 }
 FREE_IF_NON_ZERO (ppiOrig) ;
 if(ppiCount)
 {
  int i=0;
  for(i=0;i<iArraySize;i++)
  FREE_IF_NON_ZERO(ppiCount[i])
 }
 FREE_IF_NON_ZERO (ppiCount) ;
 if(pdMem)
 dMaxMemCount=max(dMaxMemCount,*pdMem);// to look at in debugger
 *piMaxRes=iRes;
 return hPaletteIs ;
}
